#  This file is part of OctoBot-Script (https://github.com/Drakkar-Software/OctoBot-Script)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot-Script. If not, see <https://www.gnu.org/licenses/>.

import os
import json
import appdirs

import octobot_script
import octobot_script.constants as constants
import octobot_script.internal.backtester_trading_mode
import octobot_commons.constants as commons_constants
import octobot_commons.logging as commons_logging
import octobot_commons.profiles as commons_profiles
import octobot_commons.tentacles_management.class_inspector as class_inspector
import octobot_tentacles_manager.api as octobot_tentacles_manager_api
import octobot_tentacles_manager.constants as octobot_tentacles_manager_constants
import octobot.configuration_manager as octobot_configuration_manager
import octobot_services.service_feeds as services_feeds
import octobot_services.api as services_api

LOGGER = commons_logging.get_logger("OctoBot-Script Tentacles")


def get_tentacles_config(forced_tentacles_by_topic=None, profile_id=None, activate_strategy_tentacles=False):
    _validate_tentacles_config_source(forced_tentacles_by_topic, profile_id)
    # use tentacles config from user appdirs as it is kept up to date at each tentacle packages install
    tentacles_setup_config = octobot_tentacles_manager_api.get_tentacles_setup_config(
        _get_tentacles_config_path(profile_id)
    )
    if not tentacles_setup_config.is_successfully_loaded:
        # reference config not available (tentacles not yet installed via CLI),
        # populate from currently imported tentacles
        octobot_tentacles_manager_api.fill_with_installed_tentacles(
            tentacles_setup_config,
            tentacles_folder=get_imported_tentacles_path()
        )
    # activate OctoBot-Script required tentacles
    _force_tentacles_config_activation(
        tentacles_setup_config,
        forced_tentacles_by_topic,
        activate_strategy_tentacles=activate_strategy_tentacles,
    )
    return tentacles_setup_config


def get_config():
    with open(get_module_config_path("config_mock.json")) as f:
        return json.load(f)


def get_module_install_path():
    return os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


def get_module_config_path(file_name):
    return os.path.join(get_module_install_path(), constants.CONFIG_PATH, file_name)


def get_module_appdir_path():
    dirs = appdirs.AppDirs(octobot_script.PROJECT_NAME, octobot_script.AUTHOR, octobot_script.VERSION)
    return dirs.user_data_dir


def get_internal_import_path():
    return os.path.join(get_module_appdir_path(), constants.ADDITIONAL_IMPORT_PATH)


def get_tentacles_path():
    return os.path.join(get_internal_import_path(), octobot_tentacles_manager_constants.TENTACLES_PATH)


def get_imported_tentacles_path():
    import tentacles
    return os.path.dirname(os.path.abspath(tentacles.__file__))


def get_public_tentacles_urls():
    return [
        octobot_configuration_manager.get_default_tentacles_url()
    ]


def force_tentacles_config_activation(
    tentacles_setup_config, forced_tentacles_by_topic=None, activate_strategy_tentacles=False
):
    _force_tentacles_config_activation(
        tentacles_setup_config,
        forced_tentacles_by_topic,
        activate_strategy_tentacles=activate_strategy_tentacles,
    )


def get_activated_social_services(forced_tentacles_by_topic=None, profile_id=None, requested_sources=None):
    social_services = []
    services_config = get_tentacles_config(forced_tentacles_by_topic, profile_id).tentacles_activation.get(
        octobot_tentacles_manager_constants.TENTACLES_SERVICES_PATH, {}
    )
    requested_sources = {str(source).lower() for source in requested_sources} if requested_sources else set()
    backtestable_feed_by_name = {
        feed_class.get_name(): feed_class
        for feed_class in services_api.get_available_backtestable_feeds()
    }
    for feed_class in class_inspector.get_all_classes_from_parent(services_feeds.AbstractServiceFeed):
        if class_inspector.is_abstract_using_inspection_and_class_naming(feed_class):
            continue
        feed_name = feed_class.get_name()
        if not services_config.get(feed_name, False):
            continue
        backtestable_feed = backtestable_feed_by_name.get(feed_name)
        if backtestable_feed is None:
            continue
        if requested_sources:
            supported_sources = {str(source).lower() for source in (backtestable_feed.get_historical_sources() or [])}
            if not requested_sources.intersection(supported_sources):
                continue
        social_services.append(feed_name)
    return sorted(social_services)


def _get_tentacles_config_path(profile_id=None):
    if profile_id:
        profile = commons_profiles.Profile.load_profile(
            constants.PROFILES_PATH or os.path.join(get_module_appdir_path(), commons_constants.USER_PROFILES_FOLDER),
            profile_id,
        )
        return profile.get_tentacles_config_path()
    return os.path.join(
        get_module_appdir_path(),
        octobot_tentacles_manager_constants.USER_REFERENCE_TENTACLE_CONFIG_PATH,
        commons_constants.CONFIG_TENTACLES_FILE
    )


def _get_forced_tentacles_activation(forced_tentacles_by_topic=None, activate_strategy_tentacles=False):
    forced_tentacles = {
        octobot_tentacles_manager_constants.TENTACLES_EVALUATOR_PATH: {},
        octobot_tentacles_manager_constants.TENTACLES_TRADING_PATH: {},
        octobot_tentacles_manager_constants.TENTACLES_SERVICES_PATH: {},
    }
    if activate_strategy_tentacles:
        import tentacles.Evaluator
        forced_tentacles[octobot_tentacles_manager_constants.TENTACLES_EVALUATOR_PATH][
            tentacles.Evaluator.BlankStrategyEvaluator.get_name()
        ] = True
        forced_tentacles[octobot_tentacles_manager_constants.TENTACLES_TRADING_PATH][
            octobot_script.internal.backtester_trading_mode.BacktesterTradingMode.get_name()
        ] = True
    for topic, activations in (forced_tentacles_by_topic or {}).items():
        forced_tentacles.setdefault(topic, {})
        forced_tentacles[topic].update(activations)
    return forced_tentacles


def _force_tentacles_config_activation(
    tentacles_setup_config,
    forced_tentacles_by_topic=None,
    activate_strategy_tentacles=False,
):
    forced_tentacles = _get_forced_tentacles_activation(
        forced_tentacles_by_topic,
        activate_strategy_tentacles=activate_strategy_tentacles,
    )
    for topic, activations in forced_tentacles.items():
        tentacles_setup_config.tentacles_activation.setdefault(topic, {})
        for tentacle, activated in activations.items():
            tentacles_setup_config.tentacles_activation[topic][tentacle] = activated
    if activate_strategy_tentacles:
        _force_backtester_trading_mode(tentacles_setup_config)


def _validate_tentacles_config_source(forced_tentacles_by_topic, profile_id):
    if forced_tentacles_by_topic is not None and profile_id is not None:
        raise ValueError("Only one of tentacles_config or profile_id can be provided.")
    if forced_tentacles_by_topic is not None and not isinstance(forced_tentacles_by_topic, dict):
        raise TypeError("tentacles_config must be a dict keyed by tentacles topic.")


def _force_backtester_trading_mode(tentacles_setup_config):
    trading_path = octobot_tentacles_manager_constants.TENTACLES_TRADING_PATH
    backtester_name = octobot_script.internal.backtester_trading_mode.BacktesterTradingMode.get_name()
    current_trading_modes = tentacles_setup_config.tentacles_activation.get(trading_path, {})
    if current_trading_modes and current_trading_modes != {backtester_name: True}:
        LOGGER.warning(
            'Trading mode is overridden by strategy() function: using BacktesterTradingMode only.'
        )
    tentacles_setup_config.tentacles_activation[trading_path] = {backtester_name: True}



#  This file is part of OctoBot-Script (https://github.com/Drakkar-Software/OctoBot-Script)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot-Script. If not, see <https://www.gnu.org/licenses/>.

import datetime

import octobot_backtesting.api as backtesting_api
import octobot_commons.symbols as commons_symbols
import octobot_commons.enums as commons_enums
import octobot_trading.enums as trading_enums
import octobot_script.internal.octobot_mocks as octobot_mocks


def _validate_tentacles_source(tentacles_config, profile_id):
    if tentacles_config is not None and profile_id is not None:
        raise ValueError("Only one of tentacles_config or profile_id can be provided.")


def _ensure_ms_timestamp(timestamp):
    if timestamp is None:
        return timestamp
    if timestamp < 16737955050:  # Friday 28 May 2500 07:57:30
        return timestamp * 1000


def _yesterday_midnight_ms() -> int:
    """Return today at 00:00:00 UTC (= end of yesterday) in milliseconds.
    Used as a stable default end_timestamp so data files collected on the
    same calendar day share an identical end boundary and can be cached."""
    today_midnight = datetime.datetime.now(datetime.timezone.utc).replace(
        hour=0, minute=0, second=0, microsecond=0
    )
    return int(today_midnight.timestamp() * 1000)


def _resolve_end_timestamp_ms(end_timestamp) -> int:
    if end_timestamp is None:
        return _yesterday_midnight_ms()
    return _ensure_ms_timestamp(end_timestamp)


async def historical_data(
    symbol,
    timeframe,
    exchange="binance",
    exchange_type=trading_enums.ExchangeTypes.SPOT.value,
    start_timestamp=None,
    end_timestamp=None,
    tentacles_config=None,
    profile_id=None,
):
    _validate_tentacles_source(tentacles_config, profile_id)
    symbols = [symbol]
    time_frames = [commons_enums.TimeFrames(timeframe)]
    start_timestamp_ms = _ensure_ms_timestamp(start_timestamp)
    end_timestamp_ms = _resolve_end_timestamp_ms(end_timestamp)
    existing_file = await backtesting_api.find_matching_data_file(
        exchange_name=exchange,
        symbols=symbols,
        time_frames=time_frames,
        start_timestamp=start_timestamp_ms,
        end_timestamp=end_timestamp_ms,
    )
    if existing_file:
        return existing_file
    data_collector_instance = (
        backtesting_api.exchange_historical_data_collector_factory(
            exchange,
            trading_enums.ExchangeTypes(exchange_type),
            octobot_mocks.get_tentacles_config(
                tentacles_config, profile_id, activate_strategy_tentacles=False
            ),
            [commons_symbols.parse_symbol(symbol) for symbol in symbols],
            time_frames=time_frames,
            start_timestamp=start_timestamp_ms,
            end_timestamp=end_timestamp_ms,
        )
    )
    return await backtesting_api.initialize_and_run_data_collector(
        data_collector_instance
    )


async def social_historical_data(
    services: list[str],
    sources: list[str] | None = None,
    symbols: list[str] | None = None,
    start_timestamp=None,
    end_timestamp=None,
    tentacles_config=None,
    profile_id=None,
):
    _validate_tentacles_source(tentacles_config, profile_id)
    start_timestamp_ms = _ensure_ms_timestamp(start_timestamp)
    end_timestamp_ms = _resolve_end_timestamp_ms(end_timestamp)
    existing_file = await backtesting_api.find_matching_data_file(
        services=services,
        symbols=symbols or [],
        start_timestamp=start_timestamp_ms,
        end_timestamp=end_timestamp_ms,
    )
    if existing_file:
        return existing_file
    data_collector_instance = backtesting_api.social_historical_data_collector_factory(
        services=services,
        tentacles_setup_config=octobot_mocks.get_tentacles_config(
            tentacles_config, profile_id, activate_strategy_tentacles=False
        ),
        sources=sources,
        symbols=[commons_symbols.parse_symbol(symbol) for symbol in symbols]
        if symbols
        else None,
        start_timestamp=start_timestamp_ms,
        end_timestamp=end_timestamp_ms,
        config=octobot_mocks.get_config(),
    )
    return await backtesting_api.initialize_and_run_data_collector(
        data_collector_instance
    )


async def get_data(
    symbol,
    time_frame,
    exchange="binance",
    exchange_type=trading_enums.ExchangeTypes.SPOT.value,
    start_timestamp=None,
    end_timestamp=None,
    data_file=None,
    social_data_files: list[str] | None = None,
    social_services: list[str] | None = None,
    social_sources: list[str] | None = None,
    social_symbols: list[str] | None = None,
    tentacles_config=None,
    profile_id=None,
):
    _validate_tentacles_source(tentacles_config, profile_id)
    data_files = (
        [data_file]
        if data_file
        else [
            await historical_data(
                symbol,
                timeframe=time_frame,
                exchange=exchange,
                exchange_type=exchange_type,
                start_timestamp=start_timestamp,
                end_timestamp=end_timestamp,
                tentacles_config=tentacles_config,
                profile_id=profile_id,
            )
        ]
    )

    if social_data_files is not None:
        data_files.extend(social_data_files)
    elif (
        profile_id is not None
        or social_sources is not None
        or tentacles_config is not None
    ):
        social_services = (
            social_services
            if social_services is not None
            else octobot_mocks.get_activated_social_services(
                tentacles_config, profile_id, requested_sources=social_sources
            )
        )
    if social_services:
        for service in social_services:
            data_files.append(
                await social_historical_data(
                    [service],
                    sources=social_sources,
                    symbols=social_symbols,
                    start_timestamp=start_timestamp,
                    end_timestamp=end_timestamp,
                    tentacles_config=tentacles_config,
                    profile_id=profile_id,
                )
            )

    return await backtesting_api.create_and_init_backtest_data(
        data_files,
        octobot_mocks.get_config(),
        octobot_mocks.get_tentacles_config(
            tentacles_config, profile_id, activate_strategy_tentacles=False
        ),
        use_accurate_price_time_frame=True,
    )



#  This file is part of OctoBot-Script (https://github.com/Drakkar-Software/OctoBot-Script)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot-Script. If not, see <https://www.gnu.org/licenses/>.
import base64
import functools
import json
import os
import shutil
import threading
import webbrowser
import http.server
import socketserver

import octobot_commons.logging as logging


class BacktestReportServer:
    def __init__(
        self,
        report_file,
        report_dir,
        report_name,
        runs_root_dir,
        server_host,
        server_port,
        serve_timeout,
        history_dir,
        data_filename,
        meta_filename,
        bundle_filename,
    ):
        self.report_file = report_file
        self.report_dir = report_dir
        self.report_name = report_name
        self.runs_root_dir = runs_root_dir
        self.server_host = server_host
        self.server_port = server_port
        self.serve_timeout = serve_timeout
        self.history_dir = history_dir
        self.data_filename = data_filename
        self.meta_filename = meta_filename
        self.bundle_filename = bundle_filename
        self.logger = logging.get_logger(self.__class__.__name__)

    @staticmethod
    def _encode_path(path):
        return base64.urlsafe_b64encode(path.encode()).decode().rstrip("=")

    def _get_all_backtesting_dirs(self):
        backtesting_dirs = []
        base_dirs = set()

        base_dirs.add(self.runs_root_dir)

        if hasattr(self, "user_data_dir") and self.user_data_dir:
            base_dirs.add(self.user_data_dir)

        for base_dir in base_dirs:
            if not os.path.isdir(base_dir):
                continue
            for root, dirs, files in os.walk(base_dir):
                for d in dirs:
                    if d.startswith("backtesting_"):
                        full_path = os.path.join(root, d)
                        backtesting_dirs.append(full_path)

        if self.report_dir not in backtesting_dirs:
            backtesting_dirs.append(self.report_dir)

        return backtesting_dirs

    def _load_meta_from_dir(self, root_path):
        meta_path = os.path.join(root_path, self.meta_filename)
        bundle_path = os.path.join(root_path, self.bundle_filename)
        if os.path.isfile(meta_path):
            with open(meta_path, encoding="utf-8") as f:
                return json.load(f)
        with open(bundle_path, encoding="utf-8") as f:
            return json.load(f).get("meta", {})

    def _has_report_data(self, root_path):
        data_path = os.path.join(root_path, self.data_filename)
        meta_path = os.path.join(root_path, self.meta_filename)
        bundle_path = os.path.join(root_path, self.bundle_filename)
        return os.path.isfile(bundle_path) or (
            os.path.isfile(meta_path) and os.path.isfile(data_path)
        )

    def _collect_history_entries(self):
        entries = []
        for run_dir in self._get_all_backtesting_dirs():
            run_name = os.path.basename(run_dir)
            history_root = os.path.join(run_dir, self.history_dir)
            try:
                for history_entry in os.scandir(history_root):
                    if not history_entry.is_dir():
                        continue
                    if not self._has_report_data(history_entry.path):
                        continue
                    try:
                        meta = self._load_meta_from_dir(history_entry.path)
                        entries.append(
                            {
                                "id": self._encode_path(history_entry.path),
                                "meta": meta,
                                "path": history_entry.path,
                                "timestamp": history_entry.name,
                                "run_name": run_name,
                            }
                        )
                    except Exception:
                        pass
            except Exception:
                pass

        return sorted(entries, key=lambda item: str(item["timestamp"]), reverse=True)

    def _clear_histories(self):
        cleared_history_dirs = 0
        cleared_run_reports = 0
        try:
            for run_dir in self._get_all_backtesting_dirs():
                is_current_run = os.path.abspath(run_dir) == os.path.abspath(
                    self.report_dir
                )
                history_root = os.path.join(run_dir, self.history_dir)
                if os.path.isdir(history_root):
                    shutil.rmtree(history_root, ignore_errors=True)
                    cleared_history_dirs += 1
                if not is_current_run:
                    removed_any = False
                    for filename in (
                        self.bundle_filename,
                        self.data_filename,
                        self.meta_filename,
                    ):
                        file_path = os.path.join(run_dir, filename)
                        if os.path.isfile(file_path):
                            try:
                                os.remove(file_path)
                                removed_any = True
                            except Exception:
                                pass
                    if removed_any:
                        cleared_run_reports += 1
        except Exception:
            pass
        return cleared_history_dirs, cleared_run_reports

    def _create_handler(self):
        server = self

        class HistoryHandler(http.server.SimpleHTTPRequestHandler):
            def _send_json(self, payload, status=200):
                body = json.dumps(payload).encode()
                self.send_response(status)
                self.send_header("Content-Type", "application/json")
                self.send_header("Content-Length", str(len(body)))
                self.end_headers()
                self.wfile.write(body)

            def do_GET(self):
                path = self.path.split("?")[0]
                if path == "/history.json":
                    history = [
                        {
                            "id": entry["id"],
                            "run_name": entry["run_name"],
                            **entry["meta"],
                        }
                        for entry in server._collect_history_entries()
                    ]
                    self._send_json(history)
                    return
                if path.startswith("/history/"):
                    self._serve_history_file(path)
                    return
                super().do_GET()

            def _serve_history_file(self, path):
                parts = path.strip("/").split("/")
                if len(parts) != 3 or parts[0] != "history":
                    self.send_error(404)
                    return
                run_id, filename = parts[1], parts[2]
                if filename not in {
                    server.data_filename,
                    server.meta_filename,
                    server.bundle_filename,
                }:
                    self.send_error(404)
                    return
                # Decode the base64 path directly instead of re-scanning the filesystem,
                # which avoids any inconsistency between the history.json scan and this request.
                try:
                    padding = (4 - len(run_id) % 4) % 4
                    decoded_path = base64.urlsafe_b64decode(
                        (run_id + "=" * padding).encode()
                    ).decode("utf-8")
                except Exception:
                    self.send_error(404)
                    return
                # Security: the decoded path must be within the allowed root directories.
                abs_decoded = os.path.realpath(decoded_path)
                allowed_roots = [
                    os.path.realpath(server.runs_root_dir),
                    os.path.realpath(server.report_dir),
                ]
                if not any(abs_decoded.startswith(root) for root in allowed_roots):
                    self.send_error(404)
                    return
                file_path = os.path.join(decoded_path, filename)
                if not os.path.isfile(file_path):
                    self.send_error(404)
                    return
                with open(file_path, "rb") as f:
                    body = f.read()
                self.send_response(200)
                self.send_header("Content-Type", "application/json")
                self.send_header("Content-Length", str(len(body)))
                self.end_headers()
                self.wfile.write(body)

            def do_POST(self):
                path = self.path.split("?")[0]
                if path != "/history-clear":
                    self.send_error(404)
                    return
                cleared_history_dirs, cleared_run_reports = server._clear_histories()
                history = [
                    {"id": entry["id"], "run_name": entry["run_name"], **entry["meta"]}
                    for entry in server._collect_history_entries()
                ]
                self._send_json(
                    {
                        "cleared": cleared_history_dirs,
                        "cleared_history_dirs": cleared_history_dirs,
                        "cleared_run_reports": cleared_run_reports,
                        "history": history,
                    }
                )

            def log_request(self, *_):
                pass

            def log_error(self, *_):
                pass

        return HistoryHandler

    def serve(self):
        handler = functools.partial(self._create_handler(), directory=self.report_dir)
        try:
            with socketserver.TCPServer(("", self.server_port), handler) as httpd:
                url = f"http://{self.server_host}:{self.server_port}/{self.report_name}"
                self.logger.info(f"Serving report on: {url}")
                opened = webbrowser.open(url, new=2)
                if not opened:
                    self.logger.error(
                        "Couldn't open a browser automatically. Open the URL manually."
                    )
                server_thread = threading.Thread(
                    target=httpd.serve_forever, daemon=True
                )
                server_thread.start()
                server_thread.join(timeout=self.serve_timeout)
                httpd.shutdown()
        except Exception:
            webbrowser.open(self.report_file, new=2)



#  This file is part of OctoBot-Script (https://github.com/Drakkar-Software/OctoBot-Script)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot-Script. If not, see <https://www.gnu.org/licenses/>.

import octobot.api as octobot_api
import octobot_backtesting.api as backtesting_api
import octobot_commons.constants as commons_constants
import octobot_commons.enums as commons_enums
import octobot_commons.logging as logging

import octobot_script.model as models
import octobot_script.internal.backtester_trading_mode as backtester_trading_mode
import octobot_script.internal.octobot_mocks as octobot_mocks


async def run(backtesting_data, strategy_config,
              enable_logs=False, enable_storage=False,
              strategy_func=None, initialize_func=None,
              tentacles_config=None, profile_id=None):
    backtest_result = models.BacktestResult(backtesting_data, strategy_config)
    registered_func = _build_script(strategy_func, initialize_func)
    _register_script(registered_func, strategy_config)
    run_tentacles_config = _resolve_run_tentacles_config(
        backtesting_data,
        strategy_func,
        tentacles_config,
        profile_id,
    )
    independent_backtesting = octobot_api.create_independent_backtesting(
        backtesting_data.config,
        run_tentacles_config,
        backtesting_data.data_files,
        run_on_common_part_only=True,
        start_timestamp=None,
        end_timestamp=None,
        enable_logs=enable_logs,
        stop_when_finished=False,
        run_on_all_available_time_frames=True,
        enforce_total_databases_max_size_after_run=False,
        enable_storage=enable_storage,
        backtesting_data=backtesting_data,
    )
    await octobot_api.initialize_and_run_independent_backtesting(independent_backtesting)
    await independent_backtesting.join_backtesting_updater(None)
    await _gather_results(independent_backtesting, backtest_result)
    await octobot_api.stop_independent_backtesting(independent_backtesting)
    return backtest_result


def _resolve_run_tentacles_config(backtesting_data, strategy_func, tentacles_config, profile_id):
    activate_strategy_tentacles = strategy_func is not None
    if tentacles_config is None and profile_id is None:
        run_tentacles_config = backtesting_data.tentacles_config
        octobot_mocks.force_tentacles_config_activation(
            run_tentacles_config,
            activate_strategy_tentacles=activate_strategy_tentacles,
        )
        return run_tentacles_config
    return octobot_mocks.get_tentacles_config(
        tentacles_config,
        profile_id,
        activate_strategy_tentacles=activate_strategy_tentacles,
    )


async def _gather_results(independent_backtesting, backtest_result):
    backtest_result.independent_backtesting = independent_backtesting
    backtest_result.duration = backtesting_api.get_backtesting_duration(
        independent_backtesting.octobot_backtesting.backtesting
    )
    backtest_result.candles_count = sum(
        candle_manager.get_preloaded_symbol_candles_count()
        for candle_manager in backtest_result.backtesting_data.preloaded_candle_managers.values()
    )
    backtest_result.report = await independent_backtesting.get_dict_formatted_report()
    backtest_result.bot_id = independent_backtesting.octobot_backtesting.bot_id


def _build_script(strategy_func, initialize_func):
    """
    Compose strategy_func and initialize_func into a single
    async callable that the backtesting engine will invoke on every candle.

    Execution order per candle:
      1. initialize_func(ctx)  => only on the very first candle tick
      2. strategy_func(ctx)    => always (signal computation)
    """
    logger = logging.get_logger("Script Runner")
    initialized = [False]  # mutable container so the closure can mutate it

    async def _combined(ctx):
        if initialize_func is not None and not initialized[0]:
            try:
                await initialize_func(ctx)
            except Exception as err:
                logger.exception(err, True, f"Failed to execute initialization function: {err}")
            finally:
                initialized[0] = True
        try:
            if strategy_func is not None:
                await strategy_func(ctx)
        except Exception as err:
            logger.exception(err, True, f"Failed to execute strategy function: {err}")
    return _combined


def _register_script(script_func, strategy_config):
    def _local_import_scripts(self, *args):
        self._live_script = script_func
        original_reload_config = self.reload_config

        async def _local_reload_config(*args, **kwargs):
            await original_reload_config(*args, **kwargs)
            updated_config = {
                commons_constants.CONFIG_ACTIVATION_TOPICS.replace(" ", "_"):
                    commons_enums.ActivationTopics.FULL_CANDLES.value
            }
            updated_config.update(strategy_config)
            self.trading_config.update(updated_config)
        self.reload_config = _local_reload_config
    backtester_trading_mode.BacktesterTradingMode._import_scripts = _local_import_scripts



#  This file is part of OctoBot-Script (https://github.com/Drakkar-Software/OctoBot-Script)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot-Script. If not, see <https://www.gnu.org/licenses/>.

import pytest
import mock

import octobot_commons.enums as commons_enums
import octobot_trading.enums as trading_enums
import octobot_backtesting.api as backtesting_api
import octobot_script as obs
import octobot_script.api.data_fetching as data_fetching

from tests import mocked_config, TEST_CONFIG, TEST_TENTACLES_CONFIG


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_historical_data():
    with (
        mock.patch(
            "octobot_script.api.data_fetching.backtesting_api.find_matching_data_file",
            mock.AsyncMock(return_value=None),
            create=True,
        ) as find_matching_data_file_mock,
        mock.patch.object(
            backtesting_api,
            "initialize_and_run_data_collector",
            mock.AsyncMock(return_value="data"),
        ) as initialize_and_run_data_collector_mock,
    ):
        assert (
            await obs.historical_data(
                "BTC/USDT", commons_enums.TimeFrames.ONE_DAY.value
            )
            == "data"
        )
        find_matching_data_file_mock.assert_awaited_once()
        initialize_and_run_data_collector_mock.assert_awaited_once()


async def test_social_historical_data():
    with (
        mock.patch(
            "octobot_script.api.data_fetching.backtesting_api.find_matching_data_file",
            mock.AsyncMock(return_value=None),
            create=True,
        ) as find_matching_data_file_mock,
        mock.patch(
            "octobot_script.api.data_fetching.backtesting_api.social_historical_data_collector_factory",
            mock.Mock(return_value="collector"),
            create=True,
        ) as factory_mock,
        mock.patch.object(
            backtesting_api,
            "initialize_and_run_data_collector",
            mock.AsyncMock(return_value="social_data"),
        ) as initialize_and_run_data_collector_mock,
    ):
        assert (
            await obs.social_historical_data(
                ["AlternativeMeServiceFeed"],
                sources=["topic_fear_and_greed"],
                start_timestamp=1700000000,
                end_timestamp=1710000000,
            )
            == "social_data"
        )
        find_matching_data_file_mock.assert_awaited_once()
        factory_mock.assert_called_once()
        initialize_and_run_data_collector_mock.assert_awaited_once()


async def test_get_data(mocked_config):
    with (
        mock.patch.object(
            data_fetching, "historical_data", mock.AsyncMock(return_value="data")
        ) as historical_data_mock,
        mock.patch.object(
            backtesting_api,
            "create_and_init_backtest_data",
            mock.AsyncMock(return_value="backtest_data"),
        ) as create_and_init_backtest_data_mock,
    ):
        assert (
            await obs.get_data("BTC/USDT", commons_enums.TimeFrames.ONE_DAY.value)
            == "backtest_data"
        )
        historical_data_mock.assert_awaited_once_with(
            "BTC/USDT",
            timeframe=commons_enums.TimeFrames.ONE_DAY.value,
            exchange="binance",
            exchange_type=trading_enums.ExchangeTypes.SPOT.value,
            start_timestamp=None,
            end_timestamp=None,
            tentacles_config=None,
            profile_id=None,
        )
        create_and_init_backtest_data_mock.assert_awaited_once_with(
            ["data"],
            TEST_CONFIG,
            TEST_TENTACLES_CONFIG,
            use_accurate_price_time_frame=True,
        )
        historical_data_mock.reset_mock()
        create_and_init_backtest_data_mock.reset_mock()
        assert (
            await obs.get_data(
                "BTC/USDT",
                commons_enums.TimeFrames.ONE_DAY.value,
                data_file="existing_file",
            )
            == "backtest_data"
        )
        historical_data_mock.assert_not_awaited()
        create_and_init_backtest_data_mock.assert_awaited_once_with(
            ["existing_file"],
            TEST_CONFIG,
            TEST_TENTACLES_CONFIG,
            use_accurate_price_time_frame=True,
        )


async def test_get_data_with_social_services(mocked_config):
    with (
        mock.patch.object(
            data_fetching,
            "historical_data",
            mock.AsyncMock(return_value="exchange_data"),
        ) as historical_data_mock,
        mock.patch.object(
            data_fetching,
            "social_historical_data",
            mock.AsyncMock(return_value="social_data"),
        ) as social_historical_data_mock,
        mock.patch.object(
            backtesting_api,
            "create_and_init_backtest_data",
            mock.AsyncMock(return_value="backtest_data"),
        ) as create_and_init_backtest_data_mock,
    ):
        assert (
            await obs.get_data(
                "BTC/USDT",
                commons_enums.TimeFrames.ONE_DAY.value,
                social_services=["CoindeskServiceFeed"],
                social_sources=["news"],
                start_timestamp=1700000000,
                end_timestamp=1710000000,
            )
            == "backtest_data"
        )
        historical_data_mock.assert_awaited_once()
        social_historical_data_mock.assert_awaited_once_with(
            ["CoindeskServiceFeed"],
            sources=["news"],
            symbols=None,
            start_timestamp=1700000000,
            end_timestamp=1710000000,
            tentacles_config=None,
            profile_id=None,
        )
        create_and_init_backtest_data_mock.assert_awaited_once_with(
            ["exchange_data", "social_data"],
            TEST_CONFIG,
            TEST_TENTACLES_CONFIG,
            use_accurate_price_time_frame=True,
        )


async def test_get_data_with_social_data_files(mocked_config):
    with (
        mock.patch.object(
            data_fetching,
            "historical_data",
            mock.AsyncMock(return_value="exchange_data"),
        ) as historical_data_mock,
        mock.patch.object(
            data_fetching,
            "social_historical_data",
            mock.AsyncMock(return_value="social_data"),
        ) as social_historical_data_mock,
        mock.patch.object(
            backtesting_api,
            "create_and_init_backtest_data",
            mock.AsyncMock(return_value="backtest_data"),
        ) as create_and_init_backtest_data_mock,
    ):
        assert (
            await obs.get_data(
                "BTC/USDT",
                commons_enums.TimeFrames.ONE_DAY.value,
                social_data_files=["social_file_1", "social_file_2"],
            )
            == "backtest_data"
        )
        historical_data_mock.assert_awaited_once()
        social_historical_data_mock.assert_not_awaited()
        create_and_init_backtest_data_mock.assert_awaited_once_with(
            ["exchange_data", "social_file_1", "social_file_2"],
            TEST_CONFIG,
            TEST_TENTACLES_CONFIG,
            use_accurate_price_time_frame=True,
        )


async def test_get_data_with_auto_social_services(mocked_config):
    with (
        mock.patch.object(
            data_fetching,
            "historical_data",
            mock.AsyncMock(return_value="exchange_data"),
        ) as historical_data_mock,
        mock.patch.object(
            data_fetching,
            "social_historical_data",
            mock.AsyncMock(return_value="social_data"),
        ) as social_historical_data_mock,
        mock.patch.object(
            data_fetching.octobot_mocks,
            "get_activated_social_services",
            mock.Mock(return_value=["AlternativeMeServiceFeed"]),
        ) as get_activated_social_services_mock,
        mock.patch.object(
            backtesting_api,
            "create_and_init_backtest_data",
            mock.AsyncMock(return_value="backtest_data"),
        ) as create_and_init_backtest_data_mock,
    ):
        assert (
            await obs.get_data(
                "BTC/USDT",
                commons_enums.TimeFrames.ONE_DAY.value,
                social_sources=["topic_fear_and_greed"],
            )
            == "backtest_data"
        )
        historical_data_mock.assert_awaited_once()
        get_activated_social_services_mock.assert_called_once_with(
            None, None, requested_sources=["topic_fear_and_greed"]
        )
        social_historical_data_mock.assert_awaited_once_with(
            ["AlternativeMeServiceFeed"],
            sources=["topic_fear_and_greed"],
            symbols=None,
            start_timestamp=None,
            end_timestamp=None,
            tentacles_config=None,
            profile_id=None,
        )
        create_and_init_backtest_data_mock.assert_awaited_once_with(
            ["exchange_data", "social_data"],
            TEST_CONFIG,
            TEST_TENTACLES_CONFIG,
            use_accurate_price_time_frame=True,
        )


async def test_get_data_with_tentacles_config(mocked_config):
    tentacles_config = {"services": {"AlternativeMeServiceFeed": True}}
    with (
        mock.patch.object(
            data_fetching,
            "historical_data",
            mock.AsyncMock(return_value="exchange_data"),
        ) as historical_data_mock,
        mock.patch.object(
            data_fetching.octobot_mocks,
            "get_activated_social_services",
            mock.Mock(return_value=[]),
        ) as get_activated_social_services_mock,
        mock.patch.object(
            data_fetching.octobot_mocks,
            "get_tentacles_config",
            mock.Mock(return_value=TEST_TENTACLES_CONFIG),
        ) as get_tentacles_config_mock,
        mock.patch.object(
            backtesting_api,
            "create_and_init_backtest_data",
            mock.AsyncMock(return_value="backtest_data"),
        ) as create_and_init_backtest_data_mock,
    ):
        assert (
            await obs.get_data(
                "BTC/USDT",
                commons_enums.TimeFrames.ONE_DAY.value,
                tentacles_config=tentacles_config,
            )
            == "backtest_data"
        )
        historical_data_mock.assert_awaited_once_with(
            "BTC/USDT",
            timeframe=commons_enums.TimeFrames.ONE_DAY.value,
            exchange="binance",
            exchange_type=trading_enums.ExchangeTypes.SPOT.value,
            start_timestamp=None,
            end_timestamp=None,
            tentacles_config=tentacles_config,
            profile_id=None,
        )
        get_activated_social_services_mock.assert_called_once_with(
            tentacles_config, None, requested_sources=None
        )
        get_tentacles_config_mock.assert_called_once_with(
            tentacles_config, None, activate_strategy_tentacles=False
        )
        create_and_init_backtest_data_mock.assert_awaited_once_with(
            ["exchange_data"],
            TEST_CONFIG,
            TEST_TENTACLES_CONFIG,
            use_accurate_price_time_frame=True,
        )


async def test_get_data_with_profile_id(mocked_config):
    profile_id = "profile-1"
    with (
        mock.patch.object(
            data_fetching,
            "historical_data",
            mock.AsyncMock(return_value="exchange_data"),
        ) as historical_data_mock,
        mock.patch.object(
            data_fetching.octobot_mocks,
            "get_activated_social_services",
            mock.Mock(return_value=[]),
        ) as get_activated_social_services_mock,
        mock.patch.object(
            data_fetching.octobot_mocks,
            "get_tentacles_config",
            mock.Mock(return_value=TEST_TENTACLES_CONFIG),
        ) as get_tentacles_config_mock,
        mock.patch.object(
            backtesting_api,
            "create_and_init_backtest_data",
            mock.AsyncMock(return_value="backtest_data"),
        ) as create_and_init_backtest_data_mock,
    ):
        assert (
            await obs.get_data(
                "BTC/USDT",
                commons_enums.TimeFrames.ONE_DAY.value,
                profile_id=profile_id,
            )
            == "backtest_data"
        )
        historical_data_mock.assert_awaited_once_with(
            "BTC/USDT",
            timeframe=commons_enums.TimeFrames.ONE_DAY.value,
            exchange="binance",
            exchange_type=trading_enums.ExchangeTypes.SPOT.value,
            start_timestamp=None,
            end_timestamp=None,
            tentacles_config=None,
            profile_id=profile_id,
        )
        get_activated_social_services_mock.assert_called_once_with(
            None, profile_id, requested_sources=None
        )
        get_tentacles_config_mock.assert_called_once_with(
            None, profile_id, activate_strategy_tentacles=False
        )
        create_and_init_backtest_data_mock.assert_awaited_once_with(
            ["exchange_data"],
            TEST_CONFIG,
            TEST_TENTACLES_CONFIG,
            use_accurate_price_time_frame=True,
        )


async def test_get_data_with_profile_id_and_tentacles_config_raises():
    with pytest.raises(ValueError):
        await obs.get_data(
            "BTC/USDT",
            commons_enums.TimeFrames.ONE_DAY.value,
            tentacles_config={"any": {}},
            profile_id="profile-1",
        )



# pylint: disable=E1101
#  This file is part of OctoBot-Script (https://github.com/Drakkar-Software/OctoBot-Script)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot-Script. If not, see <https://www.gnu.org/licenses/>.


async def plot_indicator(ctx, name, x, y, signals=None):
    # lazy import
    import octobot_script as obs

    await obs.plot(ctx, name, x=list(x), y=list(y))
    value_by_x = {x: y for x, y in zip(x, y)}
    if signals:
        await obs.plot(
            ctx,
            "signals",
            x=list(signals),
            y=[value_by_x[x] for x in signals],
            mode="markers",
        )


async def generate_and_show_report(res):
    """Generate a backtest report and show it.

    Args:
        res: BacktestResult instance

    Returns:
        The generated report object
    """
    report = await res.plot(show=False)
    print(f"Report generated at: {report.report_file}")
    report.show()
    return report



#  This file is part of OctoBot-Script (https://github.com/Drakkar-Software/OctoBot-Script)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot-Script. If not, see <https://www.gnu.org/licenses/>.
import html
import os
import re
import shutil
import time
import json

import octobot_commons.constants as commons_constants
import octobot_commons.display as display
import octobot_commons.logging as logging
import octobot_commons.timestamp_util as timestamp_util
import octobot.api as octobot_api
import octobot_tentacles_manager.api as tentacles_manager_api
import octobot_script.resources as resources
import octobot_script.internal.backtester_trading_mode as backtester_trading_mode
from octobot_script.model.backtest_report_server import BacktestReportServer


class BacktestPlot:
    DEFAULT_REPORT_NAME = "report.html"
    ADVANCED_TEMPLATE = os.path.join("dist", "index.html")
    DEFAULT_TEMPLATE = ADVANCED_TEMPLATE
    REPORT_DATA_FILENAME = "report_data.json"
    REPORT_META_FILENAME = "report_meta.json"
    REPORT_BUNDLE_FILENAME = "report.json"
    HISTORY_DIR = "backtesting"
    HISTORY_TIMESTAMP_FORMAT = "%Y%m%d_%H%M%S"
    GENERATED_TIME_FORMAT = "%Y-%m-%d at %H:%M:%S"
    SERVER_PORT = 5555
    SERVER_HOST = "localhost"
    SERVE_TIMEOUT = 300  # seconds — keep server alive for history browsing

    def __init__(self, backtest_result, run_db_identifier, report_file=None):
        self.backtest_result = backtest_result
        self.report_file = report_file
        self.run_db_identifier = run_db_identifier
        self.backtesting_analysis_settings = (
            self.default_backtesting_analysis_settings()
        )

    async def fill(self, template_file=None):
        template_name = template_file or self.DEFAULT_TEMPLATE
        template_data = await self._get_template_data()
        report_dir = os.path.dirname(os.path.abspath(self.report_file))
        try:
            shutil.copy2(
                resources.get_report_resource_path(template_name), self.report_file
            )
        except FileNotFoundError:
            if not os.path.exists(resources.get_report_resource_path(template_name)):
                raise FileNotFoundError(f"Missing report template. Please generate report templates.")
        meta = template_data["meta"]
        with open(
            os.path.join(report_dir, self.REPORT_DATA_FILENAME), "w", encoding="utf-8"
        ) as f:
            f.write(template_data["full_data"])
        with open(
            os.path.join(report_dir, self.REPORT_META_FILENAME), "w", encoding="utf-8"
        ) as f:
            json.dump(meta, f)
        bundle = {"meta": meta, "data": json.loads(template_data["full_data"])}
        with open(
            os.path.join(report_dir, self.REPORT_BUNDLE_FILENAME), "w", encoding="utf-8"
        ) as f:
            json.dump(bundle, f)
        self._save_history_entry(report_dir, template_data["full_data"], meta, bundle)

    def _save_history_entry(self, report_dir, data_str, meta, bundle):
        ts = time.strftime(self.HISTORY_TIMESTAMP_FORMAT)
        run_dir = os.path.join(report_dir, self.HISTORY_DIR, ts)
        os.makedirs(run_dir, exist_ok=True)
        with open(
            os.path.join(run_dir, self.REPORT_DATA_FILENAME), "w", encoding="utf-8"
        ) as f:
            f.write(data_str)
        with open(
            os.path.join(run_dir, self.REPORT_META_FILENAME), "w", encoding="utf-8"
        ) as f:
            json.dump(meta, f)
        with open(
            os.path.join(run_dir, self.REPORT_BUNDLE_FILENAME), "w", encoding="utf-8"
        ) as f:
            json.dump(bundle, f)

    def show(self):
        import octobot_script.internal.octobot_mocks as octobot_mocks

        report_dir = os.path.dirname(os.path.abspath(self.report_file))
        report_name = os.path.basename(self.report_file)
        runs_root_dir = os.path.dirname(report_dir)
        user_data_dir = octobot_mocks.get_module_appdir_path()
        server = BacktestReportServer(
            report_file=self.report_file,
            report_dir=report_dir,
            report_name=report_name,
            runs_root_dir=runs_root_dir,
            server_host=self.SERVER_HOST,
            server_port=self.SERVER_PORT,
            serve_timeout=self.SERVE_TIMEOUT,
            history_dir=self.HISTORY_DIR,
            data_filename=self.REPORT_DATA_FILENAME,
            meta_filename=self.REPORT_META_FILENAME,
            bundle_filename=self.REPORT_BUNDLE_FILENAME,
        )
        server.user_data_dir = user_data_dir
        server.serve()

    async def _get_template_data(self):
        full_data, symbols, time_frames, exchanges = await self._get_full_data()
        summary = self._extract_summary_metrics(full_data)
        trading_mode = self._resolve_trading_mode_class()
        return {
            "full_data": full_data,
            "meta": {
                "title": f"{', '.join(symbols)}",
                "creation_time": timestamp_util.convert_timestamp_to_datetime(
                    time.time(), self.GENERATED_TIME_FORMAT, local_timezone=True
                ),
                "strategy_config": self.backtest_result.strategy_config,
                "symbols": symbols,
                "time_frames": time_frames,
                "exchanges": exchanges,
                "summary": summary,
                "trading_mode": trading_mode.get_name() if trading_mode else None,
            },
        }

    @staticmethod
    def _extract_summary_metrics(full_data):
        try:
            parsed = json.loads(full_data)
            sub_elements = parsed.get("data", {}).get("sub_elements", [])
            details = next(
                (
                    element
                    for element in sub_elements
                    if element.get("name") == "backtesting-details"
                    and element.get("type") == "value"
                ),
                None,
            )
            values = details.get("data", {}).get("elements", []) if details else []
            metrics = {}
            for element in values:
                title = element.get("title")
                value = element.get("value")
                if title and value is not None:
                    metrics[str(title)] = str(value)
                metric_html = element.get("html")
                if metric_html:
                    metrics.update(BacktestPlot._extract_metrics_from_html(metric_html))

            def _find_metric(candidates):
                for key, value in metrics.items():
                    lowered = key.lower()
                    if any(candidate in lowered for candidate in candidates):
                        return value
                return None

            return {
                "profitability": _find_metric(
                    ("usdt gains", "profitability", "profit", "roi", "return")
                ),
                "portfolio": _find_metric(
                    ("end portfolio usdt value", "end portfolio", "portfolio")
                ),
                "metrics": metrics,
            }
        except Exception:
            return {"profitability": None, "portfolio": None, "metrics": {}}

    @staticmethod
    def _extract_metrics_from_html(metric_html):
        metrics = {}
        matches = re.findall(
            r"backtesting-run-container-values-label[^>]*>\s*(.*?)\s*</div>\s*"
            r"<div[^>]*backtesting-run-container-values-value[^>]*>\s*(.*?)\s*</div>",
            metric_html,
            flags=re.IGNORECASE | re.DOTALL,
        )
        for raw_label, raw_value in matches:
            label = BacktestPlot._html_to_text(raw_label)
            value = BacktestPlot._html_to_text(raw_value)
            if label and value:
                metrics[label] = value
        return metrics

    @staticmethod
    def _html_to_text(content):
        no_tags = re.sub(r"<[^>]+>", " ", content or "")
        return re.sub(r"\s+", " ", html.unescape(no_tags)).strip()

    async def _get_full_data(self):
        # tentacles not available during first install
        import tentacles.Meta.Keywords.scripting_library as scripting_library

        logger = logging.get_logger(self.__class__.__name__)
        elements = display.display_translator_factory()
        trading_mode = self._resolve_trading_mode_class()
        symbols = []
        time_frames = []
        exchanges = []
        for (
            exchange,
            available_symbols,
        ) in octobot_api.get_independent_backtesting_symbols_by_exchanges(
            self.backtest_result.independent_backtesting
        ).items():
            exchanges.append(exchange)
            for symbol in available_symbols:
                symbol = str(symbol)
                symbols.append(symbol)
                for time_frame in octobot_api.get_independent_backtesting_config(
                    self.backtest_result.independent_backtesting
                )[commons_constants.CONFIG_TIME_FRAME]:
                    time_frames.append(time_frame.value)
                    await elements.fill_from_database(
                        trading_mode,
                        self.run_db_identifier,
                        exchange,
                        symbol,
                        time_frame.value,
                        None,
                        with_inputs=False,
                    )
                    ctx = scripting_library.Context.minimal(
                        trading_mode,
                        logger,
                        exchange,
                        symbol,
                        self.run_db_identifier.backtesting_id,
                        self.run_db_identifier.optimizer_id,
                        self.run_db_identifier.optimization_campaign_name,
                        self.backtesting_analysis_settings,
                    )
                    try:
                        elements.add_parts_from_other(
                            await scripting_library.default_backtesting_analysis_script(
                                ctx
                            )
                        )
                    except Exception as err:
                        logger.error(
                            f"Failed to build advanced analysis for {exchange} {symbol} {time_frame.value}: {err}"
                        )
        return json.dumps(elements.to_json()), symbols, time_frames, exchanges

    def _resolve_trading_mode_class(self):
        trading_mode_name = getattr(self.run_db_identifier, "tentacle_class", None)
        if trading_mode_name:
            try:
                if (
                    trading_mode_name
                    == backtester_trading_mode.BacktesterTradingMode.__name__
                ):
                    return backtester_trading_mode.BacktesterTradingMode
                resolved = tentacles_manager_api.get_tentacle_class_from_string(
                    trading_mode_name
                )
                if resolved is not None:
                    return resolved
            except Exception:
                pass
        return backtester_trading_mode.BacktesterTradingMode

    def default_backtesting_analysis_settings(self):
        return {
            "display_backtest_details": True,
            "display_trades_and_positions": True,
            "plot_best_case_growth_on_backtesting_chart": False,
            "plot_funding_fees_on_backtesting_chart": False,
            "plot_hist_portfolio_on_backtesting_chart": True,
            "plot_pnl_on_backtesting_chart": False,
            "plot_pnl_on_main_chart": False,
            "plot_trade_gains_on_backtesting_chart": False,
            "plot_trade_gains_on_main_chart": False,
            "plot_win_rate_on_backtesting_chart": False,
            "plot_wins_and_losses_count_on_backtesting_chart": False,
            "display_backtest_details_general": False,
            "display_backtest_details_performances": True,
            "display_backtest_details_details": False,
            "display_backtest_details_strategy_settings": False,
        }



import asyncio
import os
import subprocess
import sys

from setuptools import find_packages
from setuptools import setup
from setuptools.command.build_py import build_py
from distutils.command.install import install


# from octobot_script import PROJECT_NAME, VERSION
# todo figure out how not to import octobot_script.__init__.py here
PROJECT_NAME = "OctoBot-Script"
VERSION = "0.1.1"  # major.minor.revision

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
REPORT_DIST = os.path.join(
    ROOT_DIR, "octobot_script", "resources", "report", "dist", "index.html"
)


def _build_report():
    npm = "npm"

    print("[report] Installing JS dependencies…")
    result = subprocess.run(
        [npm, "install", "--prefer-offline"],
        cwd=ROOT_DIR,
        check=False,
    )
    if result.returncode != 0:
        print("[report] 'npm install' failed – skipping JS build.", file=sys.stderr)
        return

    print("[report] Building React bundle…")
    result = subprocess.run(
        [npm, "run", "build"],
        cwd=ROOT_DIR,
        check=False,
    )
    if result.returncode != 0:
        print("[report] 'npm run build' failed – skipping JS bundle.", file=sys.stderr)
        return

    if os.path.isfile(REPORT_DIST):
        print(f"[report] Built successfully → {REPORT_DIST}")
    else:
        print("[report] Build finished but dist/index.html not found.", file=sys.stderr)


class BuildPyWithReport(build_py):
    def run(self):
        _build_report()
        super().run()


def _post_install():
    import octobot_script.cli
    asyncio.run(octobot_script.cli.install_all_tentacles(True))


class InstallWithPostInstallAction(install):
    def run(self):
        install.run(self)
        self.execute(_post_install, (), msg="Installing OctoBot-Script tentacles")


PACKAGES = find_packages(
    exclude=[
        "tests",
        "octobot_script.imports*",
        "octobot_script.user*",
    ]
)

# long description from README file
with open('README.md', encoding='utf-8') as f:
    DESCRIPTION = f.read()

REQUIRED = open('requirements.txt').readlines()
REQUIRES_PYTHON = '>=3.8'

setup(
    name=PROJECT_NAME.lower().replace("-", "_"),
    version=VERSION,
    url='https://github.com/Drakkar-Software/OctoBot-Script',
    license='GPL-3.0',
    author='Drakkar-Software',
    author_email='contact@drakkar.software',
    description='Backtesting framework of the OctoBot Ecosystem',
    packages=PACKAGES,
    cmdclass={
        'install': InstallWithPostInstallAction,
        'build_py': BuildPyWithReport,
    },
    long_description=DESCRIPTION,
    long_description_content_type='text/markdown',
    tests_require=["pytest"],
    test_suite="tests",
    zip_safe=False,
    data_files=[],
    include_package_data=True,  # copy non python files on install
    install_requires=REQUIRED,
    python_requires=REQUIRES_PYTHON,
    entry_points={
        'console_scripts': [
            'octobot_script = octobot_script.cli:main'
        ]
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Operating System :: OS Independent',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.13',
    ],
)



#  This file is part of OctoBot-Script (https://github.com/Drakkar-Software/OctoBot-Script)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot-Script. If not, see <https://www.gnu.org/licenses/>.

import mock
import pytest
import shutil
import tulipy
import os

import octobot_script as obs
import octobot_script.resources as resources
import octobot_script.model.backtest_plot as backtest_plot
from octobot_script.api.ploting import generate_and_show_report
from tests.functionnal import one_day_btc_usdt_data


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


@pytest.fixture
def mock_backtest_report_server_serve():
    with mock.patch.object(
        backtest_plot.BacktestReportServer, "serve", mock.Mock()
    ) as serve_mock:
        yield serve_mock


async def test_precomputed_vs_iteration_rsi(
    one_day_btc_usdt_data, mock_backtest_report_server_serve
):
    # 1. pre-compute entries at first iteration only
    async def _pre_compute_update(ctx):
        if run_data["entries"] is None:
            closes = await obs.Close(ctx, max_history=True)
            times = await obs.Time(ctx, max_history=True, use_close_time=True)
            rsi_v = tulipy.rsi(closes, period=ctx.tentacle.trading_config["period"])
            delta = len(closes) - len(rsi_v)
            run_data["entries"] = {
                times[index + delta]
                for index, rsi_val in enumerate(rsi_v)
                if rsi_val < ctx.tentacle.trading_config["rsi_value_buy_threshold"]
            }
            await obs.plot_indicator(
                ctx, "RSI", times[delta:], rsi_v, run_data["entries"]
            )
        if obs.current_live_time(ctx) in run_data["entries"]:
            await obs.market(
                ctx,
                "buy",
                amount="10%",
                stop_loss_offset="-15%",
                take_profit_offset="25%",
            )

    run_data = {
        "entries": None,
    }
    config = {
        "period": 10,
        "rsi_value_buy_threshold": 28,
    }
    res = await obs.run(
        one_day_btc_usdt_data,
        config,
        strategy_func=_pre_compute_update,
        enable_logs=False,
        enable_storage=True,
    )
    # ensure run happened
    assert res.backtesting_data is not None
    assert res.strategy_config is not None
    assert res.independent_backtesting is not None
    assert res.bot_id is not None
    assert res.report["bot_report"]["profitability"]["binance"] != 0
    assert (
        res.report["bot_report"]["end_portfolio"]["binance"]
        != res.report["bot_report"]["starting_portfolio"]["binance"]
    )
    assert res.duration < 10
    assert res.candles_count == 1947
    await _check_report(res)
    mock_backtest_report_server_serve.assert_called_once()

    await _reload_backtest_data(one_day_btc_usdt_data)
    # ensure second run gives the same result
    run_data = {
        "entries": None,
    }
    res_2 = await obs.run(
        one_day_btc_usdt_data,
        config,
        strategy_func=_pre_compute_update,
        enable_logs=True,
        enable_storage=False,
    )
    assert res_2.bot_id != res.bot_id
    assert (
        res_2.report["bot_report"]["profitability"]
        == res.report["bot_report"]["profitability"]
    )
    assert (
        res_2.report["bot_report"]["end_portfolio"]["binance"]
        != res_2.report["bot_report"]["starting_portfolio"]["binance"]
    )

    await _reload_backtest_data(one_day_btc_usdt_data)
    # try with different config
    run_data = {
        "entries": None,
    }
    config = {
        "period": 10,
        "rsi_value_buy_threshold": 10,
    }
    res_3 = await obs.run(
        one_day_btc_usdt_data,
        config,
        strategy_func=_pre_compute_update,
        enable_logs=False,
        enable_storage=False,
    )
    assert res_3.bot_id is not None
    assert res_3.bot_id != res.bot_id
    assert (
        res_3.report["bot_report"]["profitability"]
        != res.report["bot_report"]["profitability"]
    )
    assert (
        res_3.report["bot_report"]["end_portfolio"]["binance"]
        != res_3.report["bot_report"]["starting_portfolio"]["binance"]
    )

    # 2. iteration computed entries at each iteration
    async def _iterations_update(ctx):
        if obs.current_live_time(ctx) != await obs.current_candle_time(
            ctx, use_close_time=True
        ):
            return
        close = await obs.Close(ctx)
        if len(close) <= ctx.tentacle.trading_config["period"]:
            return
        rsi_v = tulipy.rsi(close, period=ctx.tentacle.trading_config["period"])
        if rsi_v[-1] < ctx.tentacle.trading_config["rsi_value_buy_threshold"]:
            await obs.market(
                ctx,
                "buy",
                amount="10%",
                stop_loss_offset="-15%",
                take_profit_offset="25%",
            )

    await _reload_backtest_data(one_day_btc_usdt_data)
    res_iteration = await obs.run(
        one_day_btc_usdt_data,
        config,
        strategy_func=_iterations_update,
        enable_logs=False,
        enable_storage=False,
    )
    # same result as pre_computed with the same config
    assert (
        res_iteration.report["bot_report"]["profitability"]
        == res_3.report["bot_report"]["profitability"]
    )


async def _reload_backtest_data(backtesting_data):
    """Recreate importers after a run; OctoBot closes DB handles when the run ends."""
    await backtesting_data.initialize()
    backtesting_data.preloaded_candle_managers = {}


async def _check_report(res):
    description = res.describe()
    assert str(res.strategy_config) in description
    dist_index = resources.get_report_resource_path("dist/index.html")
    assert os.path.isfile(dist_index)
    report_result = await generate_and_show_report(res)
    assert os.path.isfile(report_result.report_file)
    shutil.rmtree(os.path.dirname(report_result.report_file))
